// Copyright (c) 2018 WSO2 Inc. (http://www.wso2.org) All Rights Reserved.
//
// WSO2 Inc. licenses this file to you under the Apache License,
// Version 2.0 (the "License"); you may not use this file except
// in compliance with the License.
// You may obtain a copy of the License at
//
// http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing,
// software distributed under the License is distributed on an
// "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
// KIND, either express or implied.  See the License for the
// specific language governing permissions and limitations
// under the License.

import ballerina/system;

# Struct which represents Kafka Producer configuration.
#
# + bootstrapServers - List of remote server endpoints of Kafka brokers.
# + acks - Number of acknowledgments.
# + compressionType - Compression type to be used for messages.
# + clientID - ID to be used for server side logging.
# + metricsRecordingLevel - Metrics recording level.
# + metricReporterClasses - Metrics reporter classes.
# + partitionerClass - Partitioner class to be used to select partition to which the message is sent.
# + interceptorClasses - Interceptor classes to be used before sending records.
# + transactionalID - Transactional ID to be used in transactional delivery.
# + bufferMemory - Total bytes of memory the producer can use to buffer records.
# + noRetries - Number of retries to resend a record.
# + batchSize - Number of records to be batched for a single request. Use 0 for no batching.
# + linger - Delay to allow other records to be batched.
# + sendBuffer - Size of the TCP send buffer (SO_SNDBUF).
# + receiveBuffer - Size of the TCP receive buffer (SO_RCVBUF).
# + maxRequestSize - The maximum size of a request in bytes.
# + reconnectBackoff - Time to wait before attempting to reconnect.
# + reconnectBackoffMax - Maximum amount of time in milliseconds to wait when reconnecting.
# + retryBackoff - Time to wait before attempting to retry a failed request.
# + maxBlock - Maximum block time which the send is blocked, when the buffer is full.
# + requestTimeout - Wait time for response of a request.
# + metadataMaxAge - Maximum time to force a refresh of metadata.
# + metricsSampleWindow - Time window for a metrics sample to computed over.
# + metricsNumSamples - Number of samples maintained to compute metrics.
# + maxInFlightRequestsPerConnection - Maximum number of unacknowledged requests on a single connection.
# + connectionsMaxIdle - Close idle connections after the number of milliseconds.
# + transactionTimeout - Timeout for transaction status update from the producer.
# + enableIdempotence - Exactly one copy of each message is written in the stream when enabled.
public type ProducerConfig record {
    string? bootstrapServers = (); // BOOTSTRAP_SERVERS_CONFIG 0
    string? acks = (); // ACKS_CONFIG 1
    string? compressionType = (); // COMPRESSION_TYPE_CONFIG 2
    string? clientID = (); // CLIENT_ID_CONFIG 3
    string? metricsRecordingLevel = (); // METRICS_RECORDING_LEVEL_CONFIG 4
    string? metricReporterClasses = (); // METRIC_REPORTER_CLASSES_CONFIG 5
    string? partitionerClass = (); // PARTITIONER_CLASS_CONFIG 6
    string? interceptorClasses = (); // INTERCEPTOR_CLASSES_CONFIG 7
    string? transactionalID = (); // TRANSACTIONAL_ID_CONFIG 8

    int bufferMemory = -1; // BUFFER_MEMORY_CONFIG 0
    int noRetries = -1; // RETRIES_CONFIG 1
    int batchSize = -1; // BATCH_SIZE_CONFIG 2
    int linger = -1; // LINGER_MS_CONFIG 3
    int sendBuffer = -1; // SEND_BUFFER_CONFIG 4
    int receiveBuffer = -1; // RECEIVE_BUFFER_CONFIG 5
    int maxRequestSize = -1; // MAX_REQUEST_SIZE_CONFIG 6
    int reconnectBackoff = -1; // RECONNECT_BACKOFF_MS_CONFIG 7
    int reconnectBackoffMax = -1; // RECONNECT_BACKOFF_MAX_MS_CONFIG  8
    int retryBackoff = -1; // RETRY_BACKOFF_MS_CONFIG 9
    int maxBlock = -1; // MAX_BLOCK_MS_CONFIG 10
    int requestTimeout = -1; // REQUEST_TIMEOUT_MS_CONFIG  11
    int metadataMaxAge = -1; // METADATA_MAX_AGE_CONFIG 12
    int metricsSampleWindow = -1; // METRICS_SAMPLE_WINDOW_MS_CONFIG 13
    int metricsNumSamples = -1; // METRICS_NUM_SAMPLES_CONFIG  14
    int maxInFlightRequestsPerConnection = -1; // MAX_IN_FLIGHT_REQUESTS_PER_CONNECTION 15
    int connectionsMaxIdle = -1; // CONNECTIONS_MAX_IDLE_MS_CONFIG 16
    int transactionTimeout = -1; // TRANSACTION_TIMEOUT_CONFIG 17

    boolean enableIdempotence = false; // ENABLE_IDEMPOTENCE_CONFIG 0
    !...
};

# Represent a Kafka producer endpoint.
#
# + producerHolder - List of producers available.
# + connectorID - Unique ID for a particular connector.
# + producerConfig - Used to store configurations related to a Kafka connection.
public type SimpleProducer client object {

    public ProducerConfig? producerConfig = ();

    public function __init(ProducerConfig config) {
        self.producerConfig = config;
        var initResult = self.init(config);
        if (initResult is error) {
            panic initResult;
        }
    }

    # Initialize the producer endpoint.
    #
    # + config - Configurations related to the endpoint.
    # + return - Error if initialization failed, none otherwise.
    extern function init(ProducerConfig config) returns error?;

    public map<any> producerHolder = {};
    public string connectorID = system:uuid();

    # Aborts ongoing transaction, if transaction is initialized.
    #
    # + return - Error if aborting the transaction failed, none otherwise.
    public remote extern function abortTransaction() returns error?;

    # Closes producer connection to the external Kafka broker.
    #
    # + return - Error if closing the producer failed, none otherwise.
    public remote extern function close() returns error?;

    # Commits consumer action which commits consumer consumed offsets to offset topic.
    #
    # + consumer - Consumer which needs offsets to be committed.
    # + return - Error if committing the consumer failed, none otherwise.
    public remote extern function commitConsumer(SimpleConsumer consumer) returns error?;

    # CommitConsumerOffsets action which commits consumer offsets in given transaction.
    #
    # + offsets - Consumer offsets to commit for given transaction.
    # + groupID - Consumer group id.
    # + return - Error if committing consumer offsets failed, none otherwise.
    public remote extern function commitConsumerOffsets(PartitionOffset[] offsets, string groupID) returns error?;

    # Flush action which flush batch of records.
    #
    # + return - Error if records couldn't be flushed, none otherwise.
    public remote extern function flushRecords() returns error?;

    # GetTopicPartitions action which returns given topic partition information.
    #
    # + topic - Topic which the partition information is given.
    # + return - Partitions for the given topic, returns error if operation fails.
    public remote extern function getTopicPartitions(string topic) returns TopicPartition[]|error;

    # Simple Send action which produce records to Kafka server.
    #
    # + value - Record contents.
    # + topic - Topic to which the record will be appended to.
    # + key - Key that will be included in the record.
    # + partition - Partition to which the record should be sent.
    # + timestamp - Timestamp of the record, in milliseconds since epoch.
    # + return - Returns error if send action fails to send data, none otherwise.
    public remote extern function send(
                                          byte[] value,
                                          string topic,
                                          byte[]? key = (),
                                          int? partition = (),
                                          int? timestamp = ()
                                      ) returns error?;

};
